from typing import Callable
from functools import wraps

import asyncio, logging

from aiosow.perpetuate import perpetuate, autofill
from aiosow.bindings import return_true
from aiosow.setup import trigger_routines

trigger_routines()


def infinite_generator(condition: Callable):
    def _generator_consumer(function: Callable):
        """Re-autofill a generator function at end of it's loop"""

        iterated = None

        @wraps(function)
        async def __generator_consumer(*args, **kwargs):
            nonlocal iterated
            while await autofill(condition, args=args, **kwargs):
                if not iterated:
                    iterated = await autofill(function, args=args, **kwargs)
                try:
                    return await iterated.__anext__()
                except StopAsyncIteration:
                    iterated = await autofill(function, args=args, **kwargs)
                    return await iterated.__anext__()

        return __generator_consumer

    return _generator_consumer


ROUTINES = []


def clear_routines():
    global ROUTINES
    ROUTINES = []


def get_routines():
    global ROUTINES
    return ROUTINES


def routine(
    frequency: int | float,
    condition: Callable = return_true,
    timeout: int = -1,
    perpetuate=True,
) -> Callable:
    """
    Specifies a function to be executed as a routine.
    **args**:
    - frequency : the frequency at which the routine should run
    - condition : to prevent the triggering
    - perpetuate : wether the result should be saved in memory
    """

    def decorator(fn: Callable) -> Callable:
        logging.info("+ routine %s", fn.__name__)
        ROUTINES.append(
            {
                "frequency": frequency,
                "timeout": timeout if timeout >= 0 else abs(frequency),
                "function": fn,
                "condition": condition,
                "perpetuate": perpetuate,
            }
        )
        return fn

    return decorator


async def consume_routines(memory):
    routines = get_routines()
    # Find the routine with the smallest remaining timeout or a timeout <= 0
    smallest_timeout_routine = min(routines, key=lambda x: x["timeout"])
    smallest_timeout = smallest_timeout_routine["timeout"]

    # Wait until the smallest timeout has elapsed
    await asyncio.sleep(smallest_timeout)

    # Execute any routines that have timed out
    for routine in routines:
        routine["timeout"] -= smallest_timeout
        routine["timeout"] = 0 if routine["timeout"] < 0 else routine["timeout"]
        if routine["timeout"] <= 0:
            condition = routine["condition"]
            function = routine["function"]
            # Check the condition before running the routine
            try:
                if await autofill(condition, args=[], memory=memory):
                    if routine["perpetuate"]:
                        await perpetuate(function, args=[], memory=memory)
                    else:
                        await autofill(function, args=[], memory=memory)
            except Exception as err:
                logging.error(err)
                if memory.get("raise", False):
                    raise (err)
            # Update the timeout value based on the frequency
            if routine["frequency"] > 0:
                routine["timeout"] = routine["frequency"]
            else:
                routines.remove(routine)


async def routine_consumer(memory):  # pragma: no cover
    while True:
        await autofill(consume_routines, memory=memory)


async def spawn_routine_consumer(memory):  # pragma: no cover
    if len(ROUTINES) > 0:
        return asyncio.create_task(routine_consumer(memory))


__all__ = ["routine", "infinite_generator"]



"""
await function(*args, **memory) -> this doesn't work because :
    - it is not generic and will break whenever a function doesn't take **kwargs
    - kwargs itself is a copy and will break reference
"""
from typing import Any, Callable, List

import inspect, logging, asyncio

ALIASES = {}


def get_aliases():  # pragma: no cover
    global ALIASES
    return ALIASES


def reset_aliases():  # pragma: no cover
    global ALIASES
    ALIASES = {}


def alias(name: str):
    """
    The alias function is a decorator that can be used to inject a value for a
    function argument that will be passed to the autofill function.

    When a function is decorated with alias, the arguments with the given name
    is aliased and its value will be replaced with the value returned by the
    aliased function when calling the autofill function. The original function
    can still be called normally, but when calling the autofill function,
    the arguments value will be determined by the registered alias.
    The registered alias function will be called and its result will be used as
    the argument value.

    **Args**:
    - name (str): The name of the argument to inject a value for.

    **Returns**:
    - Callable: A decorator that takes a function as input, registers it to have
      an injected argument with the given name, and returns the function unchanged.
    """

    def decorator(function: Callable):
        ALIASES[name] = function
        return function

    return decorator


async def fill_prototype(function: Callable, args: Any = [], **kwargs) -> Any:
    """
    The autofill function takes a callable function, args, and memory as input
    arguments, and returns the result of calling the function with autofilled
    arguments.

    **Args**:
    - function (Callable): The function to be called with autofilled arguments.
    - args (Any): A list of positional arguments to pass to the function.
    - memory (Any): A dictionary of keyword arguments to pass to the function.

    **Returns**:
    - Any: The result of calling the input function with autofilled arguments.

    **Notes**:
    - If the input function has default arguments, the corresponding values will
        be used if the input args and memory do not provide values for them.

    - If the input function has been decorated, this function will unwrap the
        original function and use its signature to determine the arguments.
    """
    memory = kwargs.get("memory", {})

    def prototype(function: Callable) -> List:
        return [
            (
                param_name,
                None if param.default is inspect.Parameter.empty else param.default,
            )
            for param_name, param in inspect.signature(function).parameters.items()
            if param.kind != inspect.Parameter.VAR_KEYWORD
            and param.kind != inspect.Parameter.VAR_POSITIONAL
        ]

    class Sentinel:
        """Used to represent element to be deleted"""

    argscopy = list(args)
    if hasattr(function, "__wrapped__"):
        given_args = args
        kws = kwargs
        name = function.__wrapped__.__name__
    else:
        name = getattr(function, "__name__", None) or str(function)
        kws = kwargs if inspect.getfullargspec(function).varkw else {}
        prot = prototype(function)
        given_args = [
            memory
            if name == "memory"
            else await autofill(ALIASES[name], args=[], memory=memory)
            if name in ALIASES
            else argscopy.pop(0)
            if len(argscopy) > 0
            else (
                memory.get(name, value)
                if not inspect.getfullargspec(function).varkw
                else Sentinel
            )
            for (name, value) in prot
        ]
        given_args = [arg for arg in given_args if arg != Sentinel]
        has_varargs = any(
            param.kind == inspect.Parameter.VAR_POSITIONAL
            for param in inspect.signature(function).parameters.values()
        )
        given_args = given_args if not has_varargs else given_args + list(argscopy)
    return (name, given_args, kws)


def get_function_representation(function):
    name = function.__name__
    if name == "<lambda>":
        name = (
            inspect.getsource(function)
            .replace("\n", "")
            .replace("\t", "")
            .replace("    ", "")
        )
    return name


async def autofill(function: Callable, args: Any = [], **kwargs) -> Any:
    name, given_args, kws = await fill_prototype(function, args=args, **kwargs)
    try:
        if kwargs.get("memory", {}).get("log_autofill", False):  # pragma: no cover
            function_representation = get_function_representation(function)
            logging.info(f" -> {function_representation}")
        result = function(*given_args, **kws)
        return await result if inspect.iscoroutine(result) else result
    except Exception as err:  # pragma: no cover
        logging.error(f"{name} : {err}")
        logging.debug(f" > error generated by calling {name}({given_args})")
        raise (err)  # `debug` needs `autofill` to raise


def make_async(function: Callable) -> Callable:
    """
    Make a synchronous function run in it's own thread
    using `run_in_executor`.

    **args**:
    - function: Callable

    **returns**:
    - Async Callable
    """

    async def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        __name__, given_args, kws = await fill_prototype(function, args=args, **kwargs)
        return await loop.run_in_executor(None, function, *given_args, **kws)

    return wrapper


__all__ = ["alias", "autofill", "make_async", "fill_prototype"]



import pytest
from aiosow.setup import initialize, setup, clear_setups, get_setups


async def my_init_function_1(foo: dict):
    foo["key_1"] = "value_1"


async def my_init_function_2(foo: dict):
    foo["key_2"] = "value_2"


async def some_task():
    return "ok"


async def my_init_function_3():
    return some_task()


@pytest.mark.asyncio
async def test_initialize():
    kwargs = {"foo": {}}
    clear_setups()
    setup(my_init_function_1)
    setup(my_init_function_2)
    setup(my_init_function_3)
    tasks = await initialize(kwargs)
    assert kwargs == {"foo": {"key_1": "value_1", "key_2": "value_2"}}
    assert len(tasks) == 1
    await tasks[0]


@pytest.mark.asyncio
async def test_duplicate_setup_declaration():
    clear_setups()
    setup(my_init_function_1)
    setup(my_init_function_1)
    assert len(get_setups()) == 1



import pytest

from aiosow.bindings import delay
from aiosow.autofill import (
    get_aliases,
    reset_aliases,
    autofill,
    alias,
    get_function_representation,
)


def my_func(__a__, __b__=2, __c__=3):
    """__ is used to prevent linter error"""


@pytest.mark.asyncio
async def test_alias():
    assert len(get_aliases()) == 0

    def test():
        return "bar"

    alias("foo")(test)
    assert len(get_aliases()) == 1
    assert get_aliases()["foo"] == test


@pytest.mark.asyncio
async def test_autofill_positional():
    # Test 1: Call a simple function with positional arguments
    def my_function(a, b, c):
        return a + b + c

    args = (1, 2)
    memory = {"c": 3}

    assert await autofill(my_function, args=args, memory=memory) == 6


@pytest.mark.asyncio
async def test_autofill_positional_and_keywords():
    # Test 2: Call a function with positional and keyword arguments
    def my_function2(a, b, c=0, d=0):
        return a + b + c + d

    args = (1, 2)
    memory = {"c": 3}

    assert await autofill(my_function2, args=args, memory=memory) == 6

    args = (1, 2)
    memory = {"c": 3, "d": 4}

    assert await autofill(my_function2, args=args, memory=memory) == 10


@pytest.mark.asyncio
async def test_autofill_keywords():
    # Test 3: Call a function with keyword arguments
    def my_function3(a, b=0, c=0):
        return a + b + c

    memory = {"a": 1, "b": 2, "c": 3}

    assert await autofill(my_function3, memory=memory) == 6

    memory = {"a": 1}

    assert await autofill(my_function3, memory=memory) == 1

    memory = {"a": 1, "c": 3}

    assert await autofill(my_function3, memory=memory) == 4


@pytest.mark.asyncio
async def test_autofill_mixxed():
    # Test 4: Call a function with a mix of positional and keyword arguments
    def my_function4(a, b, c=0, d=0):
        return a + b + c + d

    args = (1,)
    memory = {"b": 2, "c": 3}

    assert await autofill(my_function4, args=args, memory=memory) == 6

    args = (1, 2, 3)
    memory = {"d": 4}

    assert await autofill(my_function4, args=args, memory=memory) == 10


@pytest.mark.asyncio
async def test_autofill_no_arg():
    # Test 5: Call a function without arguments
    def my_function5():
        return 42

    assert await autofill(my_function5) == 42


@pytest.mark.asyncio
async def test_autofill_arbitrary():
    # Test 6: Call a function with arbitrary arguments
    def my_function6(memory, *args):
        return args, memory

    args = (1, 2)
    memory = {"a": 3, "b": 4}
    assert await autofill(my_function6, args=args, memory=memory) == (
        (1, 2),
        {"a": 3, "b": 4},
    )


@pytest.mark.asyncio
async def test_autofill_decorated():
    def my_function6(memory):
        return memory

    memory = {"a": 3, "b": 4}
    # Test 7: On a decorated function
    decorated_function = delay(0.05)(my_function6)
    assert await autofill(decorated_function, memory=memory) == ({"a": 3, "b": 4})


@pytest.mark.asyncio
async def test_autofill_aliased():
    reset_aliases()

    # Test 8: With an alias
    def my_function7(test):
        return test

    args = (1, 2)
    memory = {"test": "error"}

    def test():
        return "correct"

    alias("test")(test)
    assert await autofill(my_function7, args=args, memory=memory) == "correct"


def test_get_function_representation():
    # Test case 1: Test with a named function
    def named_function():
        pass

    assert get_function_representation(named_function) == "named_function"

    # Test case 2: Test with a lambda function
    lambda_function = lambda x: x**2
    expected_result = "lambda_function = lambda x: x**2"
    assert get_function_representation(lambda_function) == expected_result

    # Test case 3: Test with a lambda function with whitespace
    lambda_function = lambda x: x**2
    expected_result = "lambda_function = lambda x: x**2"
    assert get_function_representation(lambda_function) == expected_result



import time, asyncio
import pytest
from aiosow.bindings import (
    adapter,
    expect,
    do_raise,
    dont_raise,
    return_true,
    return_false,
    call_limit,
    chain,
    delay,
    wrap,
    each,
    wire,
    accumulator,
    read_only,
    debug,
    make_async,
    until_success,
)

from unittest.mock import Mock


@pytest.fixture
def synchronous_function():
    def my_synchronous_function():
        return True

    return my_synchronous_function


@pytest.fixture
def asynchronous_function():
    async def my_asynchronous_function():
        return True

    return my_asynchronous_function


@pytest.mark.asyncio
async def test_wire_do_unlocks_listeners():
    mock_listener = Mock()
    mock_listener.__name__ = "mock"
    do_unlock_decorator, listen_decorator = wire()
    listen_decorator(mock_listener)
    await do_unlock_decorator(lambda: 1)()
    assert mock_listener.call_count == 1


@pytest.mark.asyncio
async def test_wire_do_unlocks_with_generator():
    mock_listener = Mock()
    mock_listener.__name__ = "mock"
    do_unlock_decorator, listen_decorator = wire()
    listen_decorator(mock_listener)

    def generator():
        for i in range(3):
            yield i

    await do_unlock_decorator(generator)()
    assert mock_listener.call_count == 3


@pytest.mark.asyncio
async def test_wire_chain():
    mock_start = Mock()
    mock_start.__name__ = "mock-start"
    mock_end = Mock()
    mock_end.__name__ = "mock-end"
    a_do_unlock_on, a_on_do_unlock_do = wire()
    b_do_unlock_on, b_on_do_unlock_do = wire()

    do_unlock = a_do_unlock_on(mock_start)
    value_do_unlock = a_do_unlock_on(lambda value: value)

    a_on_do_unlock_do(b_do_unlock_on(lambda value: value))
    b_on_do_unlock_do(mock_end)

    await do_unlock()
    assert mock_start.call_count == 1
    assert mock_end.call_count == 1

    await value_do_unlock(2)
    assert mock_end.call_count == 2
    mock_end.assert_called_with(2)


@pytest.mark.asyncio
async def test_accumulator():
    mock_listener = Mock()
    mock_listener.__name__ = "mock"
    batched = accumulator(2)(mock_listener)
    await batched(1)
    assert mock_listener.call_count == 0
    await batched(2)
    assert mock_listener.call_count == 1


@pytest.mark.asyncio
async def test_accumulator_from_mem():
    mock_listener = Mock()
    mock_listener.__name__ = "mock"
    batched = accumulator(lambda memory: memory["size"])(mock_listener)
    await batched(1, memory={"size": 2})
    assert mock_listener.call_count == 0
    await batched(2, memory={"size": 2})
    assert mock_listener.call_count == 1


@pytest.mark.asyncio
async def test_delay_decorator(synchronous_function, asynchronous_function):
    delayed_synchronous = delay(seconds=0.005)(synchronous_function)
    start_time = time.monotonic()
    result = await delayed_synchronous()
    end_time = time.monotonic()
    assert result is True
    assert end_time - start_time >= 0.005  # Ensure delay of at least 0.005 seconds

    delayed_asynchronous = delay(seconds=0.005)(asynchronous_function)
    start_time = time.monotonic()
    result = await delayed_asynchronous()
    end_time = time.monotonic()
    assert result is True
    assert (
        end_time - start_time >= 0.005
    )  # Ensure asynchronous function has same behavior


@pytest.mark.asyncio
async def test_wrapper_decorator():
    assert await wrap(lambda a: {"test": a})(lambda: 2)() == {"test": 2}


@pytest.mark.asyncio
async def test_each_decorator():
    async def count():
        for i in range(3):
            yield i

    async def square(n):
        return n * n

    assert await each(count)(square)() == [0, 1, 4]


@pytest.mark.asyncio
async def test_each_without_argument():
    """This test each withtout any transformation"""

    async def square(n):
        return n * n

    assert await each()(square)([0, 1, 2]) == [0, 1, 4]


@pytest.mark.asyncio
async def test_read_only():
    getter = read_only(42)
    assert getter() == 42


@pytest.mark.asyncio
async def test_debug():
    mock_listener = Mock()
    mock_listener.__name__ = "do_unlock"

    @debug(mock_listener)
    async def raises():
        raise Exception("error")

    await raises()
    assert mock_listener.call_count == 1


@pytest.mark.asyncio
async def test_make_async():
    @make_async
    def wait(*__args__):
        time.sleep(0.2)
        return "foo"

    start_time = time.monotonic()
    results = await asyncio.gather(wait(), wait(), wait())
    end_time = time.monotonic()
    assert results == ["foo", "foo", "foo"]
    assert end_time - start_time < (0.3)


@pytest.mark.asyncio
async def test_make_async_parameters():
    @make_async
    def wait(arg):
        return arg * 2

    results = await asyncio.gather(wait(2, memory={}))
    assert results == [4]


@pytest.mark.asyncio
async def test_chain():
    def add(x):
        return x + 1

    def mul(x):
        return x * 2

    def sub(x):
        return x - 3

    def _break(__x__):
        pass

    result = await chain(add, mul, sub)(1)
    # 1 + 1 = 2 ; 2 * 2 = 4 ; 4 - 3 = 1
    assert result == 1
    result = await chain(add, _break)(1)
    assert result == None
    with pytest.raises(Exception):
        result = await chain(add, _break, add)(1)
        assert result == 2


@pytest.mark.asyncio
async def test_until_success():
    attempt = 1

    @until_success()
    async def succeed_on_second_attempt():
        nonlocal attempt
        if attempt == 1:
            attempt += 1
            raise ValueError("fail once")
        else:
            return 42

    result = await succeed_on_second_attempt()
    assert result == 42


@pytest.mark.asyncio
async def test_call_limit():
    @call_limit(1)
    async def increment(x):
        return x + 1

    start_time = time.monotonic()
    result = await increment(0)
    result = await increment(0)
    diff = time.monotonic() - start_time
    assert result == 1
    assert diff >= 1


@pytest.mark.asyncio
async def test_raiser():
    @call_limit(1)
    async def raiser():
        raise ValueError("fail")

    start_time = time.monotonic()
    with pytest.raises(Exception):
        await raiser()
    with pytest.raises(Exception):
        await raiser()
    diff = time.monotonic() - start_time
    assert diff >= 1


def test_return_true():
    assert return_true() == True


def test_return_false():
    assert return_false() == False


def test_do_raise():
    with pytest.raises(ValueError):
        do_raise(ValueError)


from unittest.mock import MagicMock


@pytest.mark.asyncio
async def test_expect():
    lock = False

    def unlock(*__args__, **__kwargs__):
        nonlocal lock
        lock = True

    do_unlock = MagicMock(side_effect=unlock)
    dont_unlock = MagicMock()

    async def raise_if_lock_is_false():
        nonlocal lock
        if not lock:
            raise ValueError
        else:
            return lock

    mocked = MagicMock(side_effect=raise_if_lock_is_false)
    patched = expect(do_unlock, retries=2)(mocked)
    result = await patched()
    assert result == True

    lock = False

    with pytest.raises(ValueError):
        mocked = MagicMock(side_effect=raise_if_lock_is_false)
        patched = expect(dont_unlock, retries=1)(mocked)
        result = await patched()

    lock = False
    mocked = MagicMock(side_effect=raise_if_lock_is_false)
    patched = expect(
        dont_unlock, retries=1, on_raise=dont_raise, condition=return_false
    )(mocked)
    result = await patched()


@pytest.mark.asyncio
async def test_adapter():
    async def resolve(item):
        return item + 1

    async def multiply_by_two(num):
        return num * 2

    # Test case 1: Test with a function that takes a single argument
    adapter_func = adapter(resolve)(multiply_by_two)
    result = await adapter_func(1)
    assert result == 4



from typing import Tuple, Callable, Any, Union

import pdb as _pdb
import asyncio
import time
import inspect
from functools import wraps

from aiosow.autofill import autofill, alias, make_async
from aiosow.options import option
from aiosow.perpetuate import on, perpetuate
from aiosow.setup import setup


def adapter(resolve: Callable) -> Callable:
    """
    [unstable implementation]
    Inverse of `wrap`. Allows to mutate the parameter given to a function
    """

    def wrapper(function: Callable) -> Callable:
        async def caller(item, **kwargs):
            value = await autofill(resolve, args=[item], **kwargs)
            result = await autofill(function, args=[value], **kwargs)
            return result

        return caller

    return wrapper


def chain(*functions):
    """Applies functions iteratively and pass each result to next function"""

    async def _chain(*args, **kwargs):
        result = None
        has_result = False
        for function in functions:
            sig = inspect.getfullargspec(function)
            result = await perpetuate(
                function, args=[result] if has_result else list(args), **kwargs
            )
            if sig.annotations.get("return") is not inspect._empty:
                has_result = True
        return result

    return _chain


def until_success():
    delay = 1

    def retry_until_success(function: Callable):
        async def call(*args, **kwargs):
            nonlocal delay
            while True:
                try:
                    result = await autofill(function, args=args, **kwargs)
                    delay = 1
                    return result
                except:
                    if delay < 1200:
                        delay = delay * 2
                    await asyncio.sleep(delay)

        return call

    return retry_until_success


def wire(condition=None, perpetual=False, pass_args=True) -> Tuple[Callable, Callable]:
    """
    Returns a tuple of two decorators: `trigger_decorator` and `listen_decorator`.

    The `trigger_decorator` decorator wraps an async function and triggers it,
    calling any functions registered  with the `listen_decorator` decorator. The
    `listen_decorator` decorator registers a function to be called whenever a
    function decorated with `trigger_decorator` is called.

    **Example**:
    ```
    wire_trigger, wire_listen = wire()

    @wire_listen
    def my_function():
        print("my_function called")

    @wire_trigger
    async def my_async_function():
        print("my_async_function called")

    await my_async_function()

    # will result with

    > my_async_function_called
    > my_function_called
    ```

    If the trigger function is a [`generator`](https://docs.python.org/3/glossary.html#term-generator), `wire` wil iterate over it
    and generate a singularized task for every element. [`generator`](https://docs.python.org/3/glossary.html#term-generator)
    are function which returns a generator iterator. It looks like a normal function except that it contains yield expressions for producing a series of values usable in a for-loop or that can be retrieved one at a time with the next() function.

    **Example**:
    ```
    wire_trigger, wire_listen = wire()

    @wire_listen
    def my_function(i):
        print(f"my_function called with {i}")

    @wire_trigger
    async def my_async_function(size):
        for i in range(size):
            yield i


    await my_async_function(2)

    # will result with

    > my_function_called with 0
    > my_function_called with 1
    ```

    **Returns**:
    - A tuple of two decorators: `trigger_decorator` and `listen_decorator`.
    """
    listeners = []
    caster = autofill if not perpetual else perpetuate

    def trigger_decorator(triggerer):
        @wraps(triggerer)
        async def call(*args, **kwargs):
            result = await autofill(triggerer, args=args, **kwargs)
            # if triggerer is a generator we need to iterate over it
            if result and (
                await autofill(condition, args=[], **kwargs) if condition else True
            ):
                if inspect.isgenerator(result):
                    tasks = []
                    for val in result:
                        tasks += [
                            asyncio.create_task(
                                caster(func, args=[val] if pass_args else [], **kwargs)
                            )
                            for func in listeners
                            if func
                        ]
                else:
                    tasks = [
                        asyncio.create_task(
                            caster(func, args=[result] if pass_args else [], **kwargs)
                        )
                        for func in listeners
                        if func
                    ]
                await asyncio.gather(*tasks)
            return result

        return call

    def listen_decorator(listener):
        listeners.append(listener)
        return listener

    return (trigger_decorator, listen_decorator)


def accumulator(size: Union[int, Callable]) -> Callable:
    """
    Batch the calls to a function. Triggers it when the bucket size is reached.
    If the size passed is a `Callable`, accumulator will call it with `memory`
    to get the size.
    """

    def decorator(function: Callable) -> Callable:
        bucket = []

        async def execute(*args, **kwargs) -> Any:
            nonlocal bucket
            if isinstance(size, Callable):
                _size = size(kwargs.get("memory", {}))
            else:
                _size = size
            bucket += args
            if len(bucket) >= _size:
                argument = bucket
                bucket = []
                return await autofill(function, args=[argument], **kwargs)

        return execute

    return decorator


def call_limit(seconds):
    """
    A decorator that limits the frequency of function calls based on the number
    of seconds specified in the decorator parameter.

    **Args**:
        - seconds (int): The minimum number of seconds that must elapse between
            function calls.

    **Returns**:
        - function: A decorated function that can only be called once every
            `seconds` seconds.
    """

    def decorator(func):
        last_called = 0

        @wraps(func)
        async def wrapper(*args, **kwargs):
            nonlocal last_called
            if time.monotonic() - last_called >= seconds:
                try:
                    result = await autofill(func, args=args, **kwargs)
                    last_called = time.monotonic()
                    return result
                except Exception as e:
                    last_called = time.monotonic()
                    raise e
            else:
                await asyncio.sleep(seconds - (time.monotonic() - last_called))
                result = await autofill(func, args=args, **kwargs)
                last_called = time.monotonic()
                return result

        return wrapper

    return decorator


def delay(seconds: float) -> Callable:
    """
    Makes sure the function takes at least `seconds` to run.
    It delays the execution of an asynchronous function by
    `seconds - exec_time(function)` where `seconds` is a fixed delay time in
    seconds and `exec_time(function)` is the time taken by the wrapped function
    to execute.

    **Args**:
    - seconds: The fixed delay time in seconds.

    **Returns**:
    - A decorator that can be used to wrap an asynchronous function.

    **Example**:
    ```
        @delay(2.5)
        async def my_function():
            # code here
    ```
    """

    def decorator(function: Callable):
        @wraps(function)
        async def wrapper(*args, **kwargs):
            start_time = time.monotonic()
            result = await autofill(function, args=args, **kwargs)
            end_time = time.monotonic()
            exec_time = end_time - start_time
            delay = max(seconds - exec_time, 0)
            await asyncio.sleep(delay)
            return result

        return wrapper

    return decorator


def wrap(wrapper_function: Callable):
    """
    Wraps the result of a function with a given `wrapper_function`.

    **Args**:
    - wrapper_function: The function to use for wrapping the result of the
        decorated function.

    **Returns**:
    - A decorator that wraps the result of a function with the given
        `wrapper_function`.
    """

    def decorator(function: Callable):
        @wraps(function)
        async def execute(*args, **kwargs):
            result = await autofill(function, args=args, **kwargs)
            return wrapper_function(result)

        return execute

    return decorator


def each(iter: Union[Callable, None] = None):
    """
    Applies a function to each item in :
        - result of iter
        or
        - first argument passed to the resulting function

    **Args**:
    - iter: The async generator to iterate over.

    **Returns**:
    - A decorator that applies a function to each item returned by the given
        `iter`.
    """

    def decorator(function: Callable):
        @wraps(function)
        async def execute(*args, **kwargs):
            tasks = []
            if iter:
                subjects = await autofill(iter, args=args, **kwargs)
                async for value in subjects:
                    tasks.append(autofill(function, args=(value,), **kwargs))
            else:
                args = list(args)
                iterated = args.pop()
                for value in iterated:
                    tasks.append(autofill(function, args=(value, args), **kwargs))
            return await asyncio.gather(*tasks)

        return execute

    return decorator


def read_only(something: Any) -> Callable:
    """
    Wraps a value in a function.

    **Example**:
    ```
    getter = read_only(2)
    getter() -> 2
    ```
    **Args**:
    - something: Any

    **Returns**:
    - getter: Callable
    """

    def getter():
        nonlocal something
        return something

    return getter


def debug(trigger: Callable[[Exception, Callable, Tuple], Any]) -> Callable:
    """
    Return a decorator that calls `trigger` when the decorated entity raises
    an error.

    **Example**:
    ```
    function_to_debug = debug(pdb)(function_to_debug)
    ```

    **`trigger` is called with**:
    - error: Exception
    - function: Callable
    - args: tuple

    **Args**:
    - triggered: Callable

    **Returns**:
    - decorator: Callable
    """

    def decorator(function: Callable) -> Callable:
        @wraps(function)
        async def execute(*args, **kwargs):
            try:
                return await autofill(function, args=args, **kwargs)
            except Exception as e:
                trigger(e, function, args)

        return execute

    return decorator


def return_true(*__args__):
    """
    Python doesn't allow lambda definition inside prototype
    This function is used as a default value for conditional parameters that
    take a function to filter behavior.
    """
    return True


def return_false(*__args__):
    """
    Python doesn't allow lambda definition inside prototype
    This function is used as a default value for conditional parameters that
    take a function to filter behavior.
    """
    return False


def do_raise(exception):
    """
    Python doesn't allow lambda definition inside prototype
    This function is used as a default value for conditional parameters that
    allow custom definition of behavior on exception.
    """

    raise exception


def dont_raise(__exception__):
    """
    Python doesn't allow lambda definition inside prototype
    This function is used as a default value for conditional parameters that
    allow custom definition of behavior on exception.
    """


def expect(
    trigger: Callable,
    condition: Callable = return_true,
    retries=float("inf"),
    on_raise=do_raise,
) -> Callable:
    """
    Decorated function is retried after perpetuation of trigger function on
    Exception.
    """

    def decorator(function: Callable) -> Callable:
        counter = 0
        delay = 1

        @wraps(function)
        async def _expect(*args, **kwargs):
            nonlocal counter, delay
            try:
                result = await autofill(function, args=args, **kwargs)
                counter = 0
                return result
            except Exception as raised_error:
                if counter >= retries:
                    on_raise(raised_error)
                elif condition(args, raised_error):
                    counter += 1
                    await asyncio.sleep(delay)
                    delay = delay * 2
                    await perpetuate(trigger, args=[], **kwargs)
                    return await _expect(*args, **kwargs)
                else:
                    on_raise(raised_error)

        return _expect

    return decorator


def pdb(*__args__, **__kwargs__):  # pragma: no cover
    """
    Launches pdb.set_trace(), utility function for `aiosow.bindings.debug`

    **Example**:
    ```
    function_to_debug = debug(pdb)(function_to_debug)
    ```
    """
    _pdb.set_trace()


__all__ = [
    "alias",
    "accumulator",
    "autofill",
    "delay",
    "debug",
    "each",
    "make_async",
    "on",
    "option",
    "pdb",
    "perpetuate",
    "read_only",
    "setup",
    "wrap",
    "wire",
]
